from shaded.chained_spectral_projector import GeneralProjectionLearner
from sklearn.datasets import load_iris
from sklearn.metrics.pairwise import euclidean_distances
import numpy as np
import pytest

iris = load_iris()
X = iris.data
y = iris.target


def mean_distances(X):
    distances = euclidean_distances(X)
    len_X = len(X)
    dim_factor = 1 / (2 ** 1 / len_X)
    return dim_factor * np.sum(distances) / (len_X ** 2 - len_X)


def mean_distances_per_group(X, y):
    classes = set(y)
    mean_group_dist = []
    for class_ in classes:
        mean_group_dist.append(mean_distances(X[y == class_]))
    return np.mean(mean_group_dist)


# TODO: Repair test
@pytest.mark.skip(reason='Need to understand this before repairing')
@pytest.mark.parametrize(
    'X, y, short_chain, long_chain',
    [
        (
            X,
            y,
            ({'type': 'lda', 'args': {'n_components': 2}},),
            (
                {'type': 'lda', 'args': {'n_components': 2}},
                {'type': 'lda', 'args': {'n_components': 1}},
            ),
        )
    ],
)
def test_GeneralProjectionLearner_lda(X, y, short_chain, long_chain):
    """
    Mostly a smoke test, with the addition of testing that the mean of the intra-group distances decreases after
    projection using lda but increases with the number of components used in lda. The means are corrected for the
    numbers of dimensions following the length of the diagonal of a simplex
    """
    # find initial mean distances:
    mean_distance = mean_distances_per_group(X, y)

    # get the mean group distances for the short chain
    gpl = GeneralProjectionLearner(chain=short_chain)
    short_projected_X = gpl.fit_transform(X, y)
    short_mean = mean_distances_per_group(short_projected_X, y)

    # get the mean group distances for the long chain
    gpl = GeneralProjectionLearner(chain=long_chain)
    long_projected_X = gpl.fit_transform(X, y)
    long_mean = mean_distances_per_group(long_projected_X, y)

    assert (
        mean_distance > long_mean > short_mean
    ), 'Something is not right about the mean distances in the projected space'


# TODO: Repair test
@pytest.mark.skip(reason='Need to understand this before repairing')
def test_general_projection_learner():
    # PCA's components seems to be orthonormal
    from shaded.chained_spectral_projector import GeneralProjectionLearner
    from sklearn.datasets import load_iris

    #
    # data = load_iris()
    # X = data['data']
    # y = data['target']

    # chain = ({'type': 'pca', 'args': {'n_components': 1}}, {'type': 'pca', 'args': {'n_components': 1}})
    #
    # gpl = GeneralProjectionLearner(chain=chain)
    # XX = gpl.fit_transform(X=X, y=y, ortho=True)
    # XXX = gpl.project(X)
    # XXXX = gpl.space_residue(X)
    #
    # from oplot.plot_data_set import scatter_and_color_according_to_y
    #
    # print(gpl.scalings)
    # scatter_and_color_according_to_y(XX, y)
    # scatter_and_color_according_to_y(XXX, y)
    # scatter_and_color_according_to_y(XXXX, y)

    chain = ({'type': 'log_spaced', 'args': {'n_components': 8, 'chunk_size': 64}},)
    X = np.random.random((100, 33))
    # X = None
    gpl = GeneralProjectionLearner(chain=chain)
    gpl.fit(X=X, ortho=False)

    # print(gpl.scalings)



from shaded.linalg_utils import null_space, projection, residue
import numpy as np
import pytest


@pytest.mark.parametrize(
    'basis', [[[1, 0, 0]], [[1, 0, 0], [0, 1, 0]], (2, 3), (100, 200)]
)
def test_null_space(basis):
    if isinstance(basis, tuple):
        basis = np.random.random(basis)
    basis = np.array(basis)
    normal = null_space(basis)
    prod = np.dot(normal, basis.T)
    zero_as_prod = np.zeros(prod.shape)
    assert np.allclose(
        prod, zero_as_prod
    ), 'The null space is not normal to the original space'


@pytest.mark.parametrize(
    'basis, vector',
    [
        ([[1, 0, 0], [0, 1, 0], [0, 0, 1]], [1, 2, 3]),
        ([[1, 0, 0], [0, 1, 0], [0, 0, 1]], [[0, 1, 2], [3, 4, 5]]),
    ],
)
def test_projection(basis, vector):
    basis = np.array(basis)
    vector = np.array(vector)
    vect_proj = projection(basis, vector)
    assert np.allclose(
        vector, vect_proj
    ), 'The projection differs from the original vector, something is not right'


@pytest.mark.parametrize(
    'X, span',
    [
        (np.random.random((100, 3)), np.array([[1, 0, 0], [0, 1, 0]])),
        (np.random.random((100, 4)), np.array([[1, 0, 0, 0]])),
    ],
)
def test_proj_plus_res(X, span):
    assert np.allclose(projection(span.T, X) + residue(span.T, X), X)


@pytest.mark.parametrize(
    'span_vectors, vectors',
    [([[1, 0], [0, 1]], [[2, 3]]), ([[1, 0, 0], [0, 1, 0], [0, 0, 1]], [[2, 3, 4]])],
)
def test_residue(span_vectors, vectors):
    span_vectors = np.array(span_vectors)
    vectors = np.array(vectors)
    assert np.allclose(residue(span_vectors, vectors), np.zeros(vectors.shape))



"""Band projection matrix"""

import warnings
import numpy as np

DFLT_N_FREQ = 1025
DFLT_FREQ_RANGE = (0, DFLT_N_FREQ)


def make_buckets(
    n_buckets=15,
    freqs_weighting=lambda x: x,
    freq_range=DFLT_FREQ_RANGE,
    non_empty_bucket=True,
    reverse=False,
):
    """
    Create greedily buckets starting by aggregating lower frequencies, when the sum of
    the frequencies values so far exceed the number of buckets created times the target
    average value for a single bucket, a new bucket is created with or without the last
    term according to which choice will be closest to the target

    :param n_buckets: final number of buckets
    :param freqs_weighting: any function assigning a non-negative value to each
                            element in the freq_range
                            If given a list, the function will be assumed
                            to be the one to one mapping the frequencies to the value in the list

    :param freq_range: the range of frequencies considered, inclusive on both ends
    :param non_empty_bucket: if set to true, all buckets will have at least one element
    :param reverse: if set to true, will start aggregating higher frequencies instead, which for an increasing weight
    function will aggregate the higher frequencies into larger bins as typical
    :return: a partition of freq_range

    >>> buckets = make_buckets(n_buckets=15, freqs_weighting=lambda x: np.log(x + 0.001), freq_range=(200, 1000))
    >>> len(buckets) == 15
    True
    >>> buckets[0][0] # the first bucket starts at the first term of freq_range
    200
    >>> buckets[-1][-1] # the last bucket ends at the last term of freq_range - 1
    999

    """

    # total number of frequencies
    low_freq = freq_range[0]
    high_freq = freq_range[1]
    n_freqs = high_freq - low_freq
    if n_freqs < n_buckets and non_empty_bucket:
        warnings.warn(
            'You asked for more buckets than the number of frequencies available, '
            'some will necessarily be empty'
        )
        non_empty_bucket = False

    # indices of the frequencies
    freq_range = list(range(low_freq, high_freq + 1))
    if reverse:
        freq_range.reverse()
    # get the value of each frequency
    if not hasattr(freqs_weighting, '__iter__'):
        freq_values = list(map(freqs_weighting, freq_range))
    else:
        freq_values = freqs_weighting
    # ideal average sum of bucket value
    target_bucket_sum = sum(freq_values) / n_buckets

    # position of the next term to consider
    position = 0
    # sum of all the existing buckets plus the current bucket under construction
    existing_bucket_sum = 0
    # list of indices for each bucket, empty at moment
    idx_bucket_list = []

    for idx in range(n_buckets):
        # the list of indices in the bucket under construction
        bucket_idx = []
        while existing_bucket_sum < target_bucket_sum * (idx + 1):
            bucket_idx.append(position)
            existing_bucket_sum += freq_values[position]
            # increment the position if we are not running out of terms
            if position < n_freqs - 1:
                position += 1
            # otherwise we stop and return all the buckets, including the one under construction
            # which is non empty by construction
            else:
                idx_bucket_list.append(bucket_idx)
                if low_freq > 0:
                    idx_bucket_list = [
                        [i + low_freq for i in l] for l in idx_bucket_list
                    ]
                return idx_bucket_list

        # if we skipped the loop above, our newly constructed bucket is empty
        # if we specified that empty buckets are not ok, we force one guy in it
        if non_empty_bucket and len(bucket_idx) == 0:
            bucket_idx.append(position)
            idx_bucket_list.append(bucket_idx)
            # increment the position if we are not running out of terms
            if position < n_freqs - 1:
                position += 1
            # otherwise we stop and return all the buckets, including the one under construction
            # which is non empty by construction
            else:
                return idx_bucket_list

        else:
            # we now have a non empty bucket, so we check what is best, including the last added term or no
            total_dif_small = abs(
                existing_bucket_sum
                - freq_values[position]
                - target_bucket_sum * (idx + 1)
            )
            total_diff_large = abs(existing_bucket_sum - target_bucket_sum * (idx + 1))

            # we don't remove the last if it is better not too or if we want non empty buckets
            # and remove it would violate that rule
            if total_diff_large < total_dif_small or (
                len(bucket_idx) < 2 and non_empty_bucket
            ):
                idx_bucket_list.append(bucket_idx)
            # otherwise remove last term
            else:
                idx_bucket_list.append(bucket_idx[:-1])
                # below is to ensure that if the last bucket is empty, we increment the position
                # effectively skipping that position. Otherwise we will end up with a list of empty lists
                if len(bucket_idx[:-1]) > 0:
                    existing_bucket_sum -= freq_values[position]
                    position -= 1
                else:
                    pass

    idx_bucket_list[-1] = freq_range[idx_bucket_list[-1][0] :]
    if low_freq > 0:
        idx_bucket_list = [[i + low_freq for i in l] for l in idx_bucket_list]

    return idx_bucket_list


def make_buckets_simple(n_bucket=10, row_weights_func=lambda x: 1 + x, n_freq=20):
    """
    :param n_bucket: int, the number of buckets to create
    :param row_weights_func: a function, non zero on the integers values from 0 to n_freq. The relative values from
                             one integer a to a + 1 will become the relative number of non zero values from row a to
                             row a + 1. Note that everything has to be rounded so the actual relationship is a bit more
                             complex. The rounding follows two rules:
                             1) at least 1 in row
                             2) the excess or lack are removed from the end, one unit per index
    :param n_freq: int, the total number of frequencies, i.e, indices in buckets
    :return: a list of list of indices, consecutive, starting at 0 and ending at n_freq

    >>> make_buckets_simple(row_weights_func=lambda x: 1+x)
    [[0], [1], [2], [3], [4, 5], [6, 7], [8, 9, 10], [11, 12, 13], [14, 15, 16], [17, 18, 19, 20]]
    """

    row_weights = [row_weights_func(i) for i in range(n_bucket)]
    total_weights = np.sum(row_weights)
    n_non_zero_entry_per_unit = n_freq / total_weights
    ideal_n_per_row = [n_non_zero_entry_per_unit * w for w in row_weights]
    actual_per_row = []
    remainder = 0
    for count in ideal_n_per_row:
        rounded_count = int(max(1, round(count)))
        remainder += count - rounded_count
        actual_per_row.append(rounded_count)
    n_to_correct = int(np.abs(remainder))
    for i in range(1, n_to_correct + 1):
        actual_per_row[-i] = int(actual_per_row[-i] + np.sign(remainder))
    buckets = []
    idx = 0
    for n_in_row in actual_per_row:
        buckets.append(list(range(idx, idx + n_in_row)))
        idx += n_in_row
    return buckets


def hertz_to_mel(hertz):
    """convert a frequency given in hertz into mel"""

    return 2595 * np.log10(1 + hertz / 700)


def mel_to_hertz(mel):
    """convert a frequency given in mel into hertz"""

    return (10 ** (mel / 2595) - 1) * 700


def buckets_conversion(buckets, conversion_function=hertz_to_mel):
    """
    Convert buckets in one scale to another through the conversion_function.
    For example, if the conversion_function is set to hertz_to_mel, the function fed with hertz buckets
    will return the equivalent mel buckets.
    :param buckets: list of list of frequencies in one scale
    :param conversion_function: a mapping from frequencies in one scale to another
    :return: the converted buckets
    """

    separators = [-1]
    for bucket in buckets:
        separators.append(max(bucket))
    new_separators = [int(round(conversion_function(sep))) for sep in separators]
    new_buckets = [
        range(new_separators[i] + 1, new_separators[i + 1] + 1)
        for i in range(len(new_separators) - 1)
        if range(new_separators[i], new_separators[i + 1]) != []
    ]
    return new_buckets


def visualize_buckets(buckets, cmap='Spectral'):
    """
    Makes a plot of the buckets using a heatmap
    :param buckets: a list of the list of indices for each bucket
    :return: a plot
    """

    import matplotlib.pyplot as plt

    colors = []
    all_indices = [item for sublist in buckets for item in sublist]
    buckets.reverse()
    min_idx = min(all_indices)
    max_idx = max(all_indices)
    aspect = 10 / (max_idx - min_idx)
    for i, bucket in enumerate(buckets):
        for repeat in range(len(bucket)):
            colors.append([1 / (i + 1)])
    plt.imshow(colors, cmap=cmap, interpolation='nearest', aspect=aspect)
    plt.axis('off')
    plt.title('high frequencies on top')
    plt.show()


def make_band_matrix_row(list_entries, row_len):
    """
    Makes a row for the spectral bucket matrix. The row is zero everywhere except on the entries of index in
    list_entries where it is 1 / len(list_entries)

    :param list_entries: the indices of non zero entries of the row
    :param row_len: the length of the row
    :return: an array of length row_len as described above

    >>> make_band_matrix_row([3], 5)
    array([0., 0., 0., 1., 0.])
    >>> make_band_matrix_row([0, 3], 5)
    array([0.5, 0. , 0. , 0.5, 0. ])

    """

    n_non_zero = len(list_entries)
    row = np.array(
        [0 if i not in list_entries else 1 / n_non_zero for i in range(row_len)]
    )
    return row


def make_band_matrix(buckets, n_freq=DFLT_N_FREQ):
    """
    Given a list of n list of indices, make a matrix of size n by n_freq where the entries of row k are 0
    everywhere except at the index in the k list of buckets, where the entries are the inverse of the length
    of that bucket

    :param buckets: a list of list containing the indices of non zero entries of the corresponding row
    :param n_freq: the number of column of the matrix
    :return: a len(buckets) by n_freq matrix

    >>> buckets = make_buckets(n_buckets=15, freqs_weighting=lambda x: np.log(x + 0.001), freq_range=(200, 1000))
    >>> M = make_band_matrix(buckets, n_freq=1025)
    >>> print(M.shape)
    (15, 1025)

    >>> # the matrix sends each of the 200 first unitary vector to the zero vector (not below is not a mathematical proof of that, but solid clue)
    >>> vec = [1] * 200 + [0] * (1025 - 200)
    >>> print(np.dot(M, vec))
    [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]

    >>> # after that, we get non zero output (in general at least)
    >>> vec = [1] * 200 + [1] + [0] * (1025 - 201)
    >>> print(np.dot(M, vec))
    [0.01612903 0.         0.         0.         0.         0.
     0.         0.         0.         0.         0.         0.
     0.         0.         0.        ]

    """

    n_bands = len(buckets)
    bucket_matrix = np.zeros((n_bands, n_freq))
    for row_idx, bucket in enumerate(buckets):
        row = make_band_matrix_row(list_entries=bucket, row_len=n_freq)
        bucket_matrix[row_idx] = row
    return bucket_matrix


def make_band_proj_matrix(
    n_buckets=15,
    n_freq=DFLT_N_FREQ,
    freqs_weighting=lambda x: x,
    non_empty_bucket=True,
    reverse=False,
):
    """
    Uses make_buckets and make_band_matrix to make a band-based projection matrix

    :param n_buckets: final number of buckets
    :param freqs_weighting: any function assigning a non-negative value to each
                            element in the freq_range
                            If given a list, the function will be assumed
                            to be the one to one mapping the frequencies to the value in the list

    :param n_freq: the number of column of the matrix
    :param non_empty_bucket: if set to true, all buckets will have at least one element
    :param reverse: if set to true, will start aggregating higher frequencies instead
    :return: a len(buckets) by n_freq matrix

    >>> make_band_proj_matrix(n_buckets=3, n_freq=10)
    array([[0.14285714, 0.14285714, 0.14285714, 0.14285714, 0.14285714,
            0.14285714, 0.14285714, 0.        , 0.        , 0.        ],
           [0.        , 0.        , 0.        , 0.        , 0.        ,
            0.        , 0.        , 0.33333333, 0.33333333, 0.33333333]])

    """
    buckets = make_buckets(
        n_buckets, freqs_weighting, (0, n_freq), non_empty_bucket, reverse
    )
    return make_band_matrix(buckets, n_freq)



"""Util"""


try:
    import importlib.resources

    _files = importlib.resources.files  # only valid in 3.9+
except AttributeError:
    import importlib_resources  # needs pip install

    _files = importlib_resources.files

files = _files('shaded')
data_dir = files / 'data'
data_dir_path = str(data_dir)



"""Chained spectral projectors"""

import importlib
import numpy as np
import warnings

from sklearn.neighbors import NeighborhoodComponentsAnalysis as NCA
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA
from sklearn.decomposition import PCA
from sklearn.linear_model import LinearRegression
from sklearn import random_projection

from shaded.pseudo_pca import PseudoPca
from shaded.pseudo_lda import PseudoLda
from shaded.pair_wise_lda import PairwiseLda
from shaded.linalg_utils import residue, orth
from shaded.band_projection_matrix import make_band_matrix, make_buckets, hertz_to_mel

CLUSTERING_OPTIONS = (
    'KMeans',
    'SpectralClustering',
    'AffinityPropagation',
    'AgglomerativeClustering',
    'Birch',
    'MeanShift',
)


def learn_spect_proj(
    X,
    y=None,
    spectral_proj_name='pca',
    clustering_meth='KMeans',
    clustering_options=CLUSTERING_OPTIONS,
    kwargs_feat=None,
    kwargs_clust=None,
):
    """
    Function to learn each of the important spectral projection

    :param X: the fvs, an array of size n*k
    :param y: the classes, an array of size n
    :param spectral_proj_name: a string of the name of the featurizer
    :param args: extra argument to be passed to the featurizer class
    :return: a matrix in the form of a numpy array
    """

    clustering_options = set(clustering_options)
    kwargs_feat = kwargs_feat or {'n_components': 10}
    kwargs_clust = kwargs_clust or {}

    assert (
        clustering_meth in clustering_options
    ), 'clustering options must one of {}'.format(
        ', '.join(map(str, clustering_options))
    )
    clusterer_m = getattr(importlib.import_module('sklearn.cluster'), clustering_meth)

    if spectral_proj_name == 'random_gaussian':
        rand_proj = random_projection.GaussianRandomProjection(**kwargs_feat)
        rand_proj.fit(X)
        proj_matrix = rand_proj.components_.T

    elif spectral_proj_name == 'keep_features':
        indices = kwargs_feat['indices']
        proj_matrix = np.zeros((X.shape[1], len(indices)))
        for idx in range(len(indices)):
            proj_matrix[indices[idx], idx] = 1

    elif spectral_proj_name == 'pca':
        pca = PCA(**kwargs_feat)
        pca.fit(X)
        proj_matrix = pca.components_.T

    elif spectral_proj_name == 'pseudo_pca':
        # make the pseudo pca proj matrix
        ppca = PseudoPca(**kwargs_feat)
        ppca.fit(X)
        proj_matrix = ppca.proj_mat.T

    elif spectral_proj_name == 'lda':
        lda = LDA(**kwargs_feat)
        lda.fit(X, y)
        n_components = kwargs_feat['n_components']
        proj_matrix = lda.scalings_[:, :n_components]

    elif spectral_proj_name == 'pseudo_lda':
        plda = PseudoLda(**kwargs_feat)
        plda.fit(X, y)
        proj_matrix = plda.proj_mat

    elif spectral_proj_name == 'pairwise_lda':
        plda = PairwiseLda(**kwargs_feat)
        classes = kwargs_feat.get('classes', None)
        normal = kwargs_feat.get('normal', False)
        plda.fit(X, y, classes=classes, normal=normal)
        proj_matrix = plda.proj_mat

    elif spectral_proj_name == 'unsupervised_lda':
        n_components = kwargs_feat['n_components']
        if y is not None:
            warning = (
                'Warning: y will be replaced by classes found by the chosen clusterer'
            )
            warnings.warn(warning)
        if 'n_clusters' in clusterer_m.__init__.__code__.co_varnames:
            y = clusterer_m(n_clusters=n_components + 1, **kwargs_clust).fit_predict(X)
        else:
            y = clusterer_m(**kwargs_clust).fit_predict(X)
        lda = LDA(**kwargs_feat)
        lda.fit(X, y)
        proj_matrix = lda.scalings_[:, :n_components]

    elif spectral_proj_name == 'unsupervised_pseudo_lda':
        n_components = kwargs_feat['n_components']
        if y is not None:
            warning = (
                'Warning: y will be replaced by classes found by the chosen clusterer'
            )
            warnings.warn(warning)
        if 'n_clusters' in clusterer_m.__init__.__code__.co_varnames:
            y = clusterer_m(n_clusters=n_components + 1, **kwargs_clust).fit_predict(X)
        else:
            y = clusterer_m(**kwargs_clust).fit_predict(X)
        pseudo_lda = PseudoLda(**kwargs_feat)
        pseudo_lda.fit(X, y)
        proj_matrix = pseudo_lda.proj_mat[:, :n_components]

    elif spectral_proj_name == 'nca':
        nca = NCA(**kwargs_feat)
        nca.fit(X, y)
        proj_matrix = nca.components_.T

    elif spectral_proj_name == 'unsupervised_nca':
        n_components = kwargs_feat['n_components']
        if y is not None:
            print('y will be replaced by classes found by the chosen clusterer')
        if 'n_clusters' in clusterer_m.__init__.__code__.co_varnames:
            y = clusterer_m(n_clusters=n_components + 1, **kwargs_clust).fit_predict(X)
        else:
            y = clusterer_m(**kwargs_clust).fit_predict(X)
        nca = NCA(**kwargs_feat)
        nca.fit(X, y)
        proj_matrix = nca.components_.T

    elif spectral_proj_name == 'linear_regression':
        lr = LinearRegression()
        lr.fit(X, y)
        proj_matrix = lr.coef_.T

    elif spectral_proj_name == 'default':
        n_components = kwargs_feat['n_components']
        X = np.array(X)
        n_freq = X.shape[1]
        freqs_weighting = kwargs_feat.get('freqs_weighting', hertz_to_mel)
        buckets = make_buckets(
            n_buckets=n_components,
            reverse=False,
            freqs_weighting=freqs_weighting,
            freq_range=(0, n_freq),
        )
        default_matrix = make_band_matrix(buckets, n_freq=n_freq)
        proj_matrix = default_matrix.T

    else:
        all_spectral_proj = ', '.join(
            [
                'keep_features',
                'pca',
                'lda',
                'pseudo_pca',
                'unsupervised_lda',
                'unsupervised_nca',
                'nca',
                'linear_regression',
                'unsupervised_pseudo_lda',
                'pseudo_lda',
                'pairwise_lda',
                'random_gaussian',
                'default',
            ]
        )
        raise ValueError(f'the spectral projector must be one of: {all_spectral_proj}')

    return proj_matrix


def keep_only_indices(indices, n_freq=1025):
    """
    Makes a matrix which when a spectra is multiplied with it, only the entires in the list indices are kept
    :param indices: a list of indices to keep
    :param n_freq: the total number of frequencies in the spectra
    :return: a matrix of size (n_freq, len(indices))
    """

    proj_matrix = np.zeros((n_freq, len(indices)))
    for idx in range(len(indices)):
        proj_matrix[indices[idx], idx] = 1
    return proj_matrix


def learn_chain_proj_matrix(
    X,
    y=None,
    chain=({'type': 'pca', 'args': {'n_components': 5}},),
    indices=None,
    n_freq=1025,
):
    """
    A function successively learning a projections matrix on the residue of the previous one. The projections
    matrices are then concatenated and return as one single projection matrix. Note that the final projection
    matrix may not produce fvs of the size the sum of the components of each part, i.e., care must be taken
    to ensure each classes called must be able to produce the required number of components. For example,
    if the number of classes is 10, then lda can only produce 9 dimensions. To obtain say 12 dimension, the user
    will need to chain two lda's, for example with size 9 and 3 respectively.

    :param X: the fvs, an array of size n*k
    :param y: the classes, an array of size n
    :param chain: a tuple of dictionaries each containing the type of projection along with its parameters
    :param indices: the indices of the spectra to work with, anything else is discarded
    :param n_freq: the total number of entries from the spectra. Only needed if n_freq is not None, in order to
                   determine the size of the freq_selection_matrix
    :return: a single projection matrix
    """

    from oplot.plot_data_set import scatter_and_color_according_to_y

    freq_selection_matrix = None
    if indices is not None:
        freq_selection_matrix = keep_only_indices(indices, n_freq=n_freq)
        X = np.dot(X, freq_selection_matrix)

    all_proj_matrices = []
    for idx, mat_dict in enumerate(chain):
        kwargs_feat = mat_dict['args']
        proj_matrix = learn_spect_proj(
            X, y, spectral_proj_name=mat_dict['type'], kwargs_feat=kwargs_feat
        )
        all_proj_matrices.append(proj_matrix)
        # if idx < len(chain) - 1:
        X = residue(proj_matrix, X)
        # # scatter_and_color_according_to_y(X, y)
        # print(X[:3])

    proj_matrix = np.hstack(tuple(all_proj_matrices))
    if indices is not None:
        proj_matrix = np.dot(freq_selection_matrix, proj_matrix)
    return np.array(proj_matrix)


from sklearn.base import TransformerMixin, BaseEstimator


class GeneralProjectionLearner(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        chain=({'type': 'pca', 'args': {'n_components': 5}},),
        indices=None,
        n_freq=1025,
    ):
        self.chain = chain
        self.indices = indices
        self.n_freq = n_freq

    def fit(self, X, y=None, ortho=True):
        """
        Ortho set to True will make it so that the projection into the subspace will be "direction agnostic", which
        makes sense in terms of pure dimension reduction. But for some model, like lda, the length of the vectors onto
        which we will project represent the relative importance of those directions. In that case it may make sense to
        set ortho to False

        :param X: array of vectors
        :param y: array of tags
        :param ortho: boolean, whether or not to orthonormalize the vector onto which to project
        :return:
        """

        self.scalings = learn_chain_proj_matrix(
            X, y, self.chain, indices=self.indices, n_freq=self.n_freq
        )
        if ortho:
            self.scalings = orth(self.scalings.T).T
            self.projection_ = np.dot(self.scalings, self.scalings.T)
        else:
            self.projection_ = np.dot(self.scalings, self.scalings.T)
        return self

    def transform(self, X):
        # lower the dimension of the vectors by projecting in the subspace
        return np.dot(X, self.scalings)

    def project(self, X):
        # projects in the SAME space, the dimension is unchanged
        return np.dot(X, self.projection_)

    def space_residue(self, X):
        return X - np.dot(X, self.projection_)



"""
Pairwise LDA
"""
from sklearn.base import BaseEstimator, TransformerMixin
import numpy as np
from itertools import combinations
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA

from shaded.linalg_utils import residue


class PairwiseLda(BaseEstimator, TransformerMixin):
    """
    A class finding the hyperplane spanning the lines found by LDA when trying to
    separate the pair of classes
    """

    def __init__(self, n_components=None):
        self.n_components = n_components

    # if normal is set to True, the projection vectors are taken in the residue of the previously found ones
    def fit(self, X, y, classes=None, normal=False):
        if classes is None:
            classes = set(y)
        scalings_ = []
        for pair_classes in combinations(classes, 2):
            pair_mask = np.isin(y, pair_classes)
            X_pair = X[pair_mask]
            y_pair = y[pair_mask]
            lda = LDA()
            lda.fit(X_pair, y_pair)
            scalings_.append(lda.scalings_.flatten())
            if normal:
                X = residue(lda.scalings_, X)
        self.proj_mat = np.array(scalings_).T

    def transform(self, X, y=None):
        return np.dot(X, self.proj_mat.T)



"""Pseudo LDA"""

from sklearn.base import BaseEstimator, TransformerMixin
import numpy as np


class PseudoLda(BaseEstimator, TransformerMixin):
    """
    A class finding the hyperplane passing the centers of the classes in y and projecting the vectors in X
    onto that hyperplane. The number of components is fixed to n_classes - 1, no more no less. If the classes form
    well separated spherical clusters, the class will work as well as LDA but be much faster.
    Trouble will arise if the clusters are parallel ellipsoides, close enough that there centers will be relatively
    close in the hyperplane chosen by PseudoLda.
    """

    def __init__(self, n_components):
        self.hyper_dim = n_components + 1

    def fit(self, X, y):
        y = np.array(y)
        X = np.array(X)
        centers = []
        for class_ in set(y):
            cluster = X[y == class_]
            centers.append(np.mean(cluster, axis=0))
        centers = np.array(centers)
        hyper_plane_basis = centers - np.tile(centers[0, :], (centers.shape[0], 1))
        hyper_plane_basis = hyper_plane_basis[1:, :]
        unit_matrix = np.array(
            [
                vec / norm
                for vec, norm in zip(
                    hyper_plane_basis, np.sum(hyper_plane_basis ** 2, axis=1) ** 0.5
                )
            ]
        )
        self.proj_mat = unit_matrix.T

    def transform(self, X, y=None):
        return np.dot(X, self.proj_mat)



"""Pseudo PCA"""

from sklearn.base import BaseEstimator, TransformerMixin
import numpy as np


class PseudoPca(BaseEstimator, TransformerMixin):
    """
    A class imitating PCA in the following way:
    """

    def __init__(self, n_components, n_samples=None):
        self.hyper_dim = n_components + 1
        if n_samples is None:
            n_samples = n_components
        self.n_samples = n_samples

    def get_random_hyperplane_unit_matrix(self, X):
        rand_points = X[np.random.choice(X.shape[0], self.hyper_dim)]
        hyper_plane_basis = rand_points - np.tile(
            rand_points[0, :], (rand_points.shape[0], 1)
        )
        hyper_plane_basis = hyper_plane_basis[1:, :]
        unit_matrix = np.array(
            [
                vec / norm
                for vec, norm in zip(
                    hyper_plane_basis, np.sum(hyper_plane_basis ** 2, axis=1) ** 0.5
                )
            ]
        )
        return unit_matrix

    def fit(self, X, y=None):
        unit_matrices = []
        i = 0
        while i <= self.n_samples:
            rand_hyper = self.get_random_hyperplane_unit_matrix(X)
            # this check is here for the following reason: rand_hyper may have nan's.
            # this can happen when in get_random_hyperplane_unit_matrix the hyper_plane_basis
            # has the first row repeated (requiring X to have repeated vectors), then hyper_plane_basis
            # can have rows of zero, in which case the division by norm in unit matrix yields some nan
            if np.isnan(np.max(rand_hyper)):
                pass
            else:
                unit_matrices.append(rand_hyper)
                i += 1
        self.proj_mat = np.mean(np.array(unit_matrices), axis=0)

    def transform(self, X, y=None):
        return np.dot(X, self.proj_mat.T)



"""
Linear algebra utils
"""
import numpy as np


def null_space(A, rcond=None):
    """
    Find an orthonormal matrix whose rows span the nullspace of the matrix A. Put another way, find a maximal
    number of row vectors, each of which has a 0 dot product with all the rows of A, while those row vectors are
    also normal to each other and of unit length

    >>> A = np.array([[1, 0, 0]])
    >>> null_space(A)
    array([[0., 1., 0.],
           [0., 0., 1.]])

    >>> A = np.array([[1, -2, 6, 8], [2.1, 6, 20, -3]])
    >>> B = null_space(A)

    The vectors are normal

    >>> np.sum(B ** 2, axis=1)
    array([1., 1.])
    """

    u, s, vh = np.linalg.svd(A, full_matrices=True)
    M, N = u.shape[0], vh.shape[1]
    if rcond is None:
        rcond = np.finfo(s.dtype).eps * max(M, N)
    tol = np.amax(s) * rcond
    num = np.sum(s > tol, dtype=int)
    Q = vh[num:, :].T.conj()
    return Q.T


def projection(basis, vectors):
    """
    The vectors live in a k dimensional space S and the columns of the basis are vectors of the same
    space spanning a subspace of S. Gives a representation of the projection of vector into the space
    spanned by basis in term of the basis.

    :param basis: an n-by-k array, a matrix whose vertical columns are the vectors of the basis
    :param vectors: an m-by-k array, a vector to be represented in the basis
    :return: an m-by-k array
    """
    # most consistent with math would be np.dot(basis.T, np.dot(basis, vector))
    return np.dot(np.dot(vectors, basis), basis.T)


def make_proj_matrix(span_vectors):
    span_vectors = orth(span_vectors)
    return np.dot(span_vectors.T, span_vectors)


def projection_(proj_mat, vectors):
    return np.dot(vectors, proj_mat)


def residue(span_vectors, vectors):
    """
    find the residue of each of vectors after projection in basis
    :param span_vectors: an n-by-k array, spanning a vector space A
    :param vectors: an n-by-l array
    :return: an n-by-l array, the residues of vectors with the respect of the projection in the space spanned
             by span_vectors

    >>> span_vectors = np.array([[1, 0], [0, 1]])
    >>> vectors = np.array([[2, 3]])
    >>> print(residue(span_vectors, vectors))
    [[0 0]]
    >>> span_vectors = np.array([[1, 0],[0, 0]])
    >>> vectors = np.array([[2, 3]])
    >>> print(residue(span_vectors, vectors))
    [[0 3]]
    """

    return vectors - projection(span_vectors, vectors)


def orth(A):
    """
    Orthonormalize the matrix A, i.e, replace it with another matrix spanning the same subspace but:
    1) with minimal number of rows
    2) rows normal to each other
    3) rows of length 1

    :param A: a n-by-k array
    :return: a n-by-k array whose row spans the same space as the rows of A and with the normal rows

    >>> M = np.array([[1,2,3], [3,4,5]])
    >>> OM = orth(M)
    >>> # checking the product of O and its transpose is the identity
    >>> np.allclose(np.dot(OM, OM.T), np.eye(2))
    True
    """

    u, s, vh = np.linalg.svd(A.T, full_matrices=False)
    N, M = A.shape
    eps = np.finfo(float).eps
    tol = max(np.max(M), np.max(N)) * np.amax(s) * eps
    num = np.sum(s > tol, dtype=int)
    Q = u[:, :num]
    return Q.T


# def relative_orth(A, B):
#     """
#     Compute an orthonormal array B', all of the rows of which are normal to A
#     and spanning the same space as B itself.
#     Note that A is not required to be orthonormal and A is left unchanged.
#
#     :param A: a n-by-k array
#     :param B: a m-by-k array
#     :return: a m-by-k array B' with the property that span(A, B') = span(A, B)
#
#     """
#     residue_B = residue(A, B)
#
#     return orth(residue_B)
